#!/usr/bin/env python
""" A multiprocessor wrapper around pngcrush.
    Copyright (C) 2010 Kevin Vance <kvance@kvance.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__version__ = '1.0'

import optparse
import os
import re
import shutil
import subprocess
import sys
import tempfile


# If python >= 2.6 is installed, we can get a useful CPU count.
try:
    import multiprocessing
    DEFAULT_NUM_JOBS = multiprocessing.cpu_count()
except ImportError:
    DEFAULT_NUM_JOBS = 1


#
# Worker class
#

class CrushJob:
    """A single pngcrush process."""
    def __init__(self, infile, methods, extra_args=None):
        """Initialize this job.

        Arguments:
        infile - Input filename to crush
        methods - List of method numbers, e.g. [1, 4, 7]
        extra_args - List of arguments to pass to pngcrush directly

        """
        self.process = None
        if extra_args is None:
            extra_args = []

        # Create a tempfile to use for output.
        tf_obj = tempfile.NamedTemporaryFile(delete=False)
        tf_obj.close()
        self.outfile = tf_obj.name

        # Build the command line.
        method_cmds = sum((['-m', str(x)] for x in methods), [])
        files = [infile, self.outfile]
        self.cmdline = ['pngcrush'] + method_cmds + extra_args + files


    def __del__(self):
        # If the process is still running, kill it.
        if self.process is not None:
            try:
                self.process.terminate()
            except OSError:
                pass

        # Clean up our temp file
        os.unlink(self.outfile)


    def start(self):
        """Start running this job in the background."""
        self.process = subprocess.Popen(self.cmdline)


    def result(self):
        """Return (filename, size) after waiting for the job to finish."""
        if self.process is None:
            raise RuntimeError("Job has not started yet")
        self.process.wait()
        return CrushResult(self.outfile)


class CrushResult:
    """The result of a CrushJob."""
    def __init__(self, path):
        self.path = path
        self.size = os.path.getsize(path)


#
# Utilities
#

def process_args():
    """Get our options, our arguments, and pngcrush's arguments from argv."""
    # Split argv at -- with our args on the left, and pngcrush's on the right.
    args = sys.argv[1:]
    try:
        split_idx = args.index('--')
    except ValueError:
        our_argv = args
        pngcrush_argv = []
    else:
        our_argv = args[:split_idx]
        pngcrush_argv = args[split_idx+1:]

    # Parse our argv into options and arguments.
    usage = 'usage: %prog [options] <input.png> <output.png> '\
            '[-- pngcrush options]'
    parser = optparse.OptionParser(usage=usage)
    parser.add_option('-j', dest='jobs', type='int', default=DEFAULT_NUM_JOBS,
                      help='How many jobs to run concurrently')
    options, args = parser.parse_args(our_argv)

    # Validate everything before returning.
    if len(args) == 0:
        parser.error("Input file is required")
    if len(args) == 1:
        parser.error("Output file is required")
    if len(args) > 2:
        parser.error("Too many arguments")
    if not os.path.exists(args[0]):
        parser.error("Input file does not exist")

    return options, args, pngcrush_argv


def split_rr(a_list, count):
    """Split list object 'a_list' into 'count' lists, round-robin.

    >>> split_rr([1, 2, 3, 4, 5, 6], 3)
    ([1, 4], [2, 5], [3, 6])

    """
    return tuple(a_list[i::count] for i in range(count))


def get_brute_range():
    """Return the range of methods to use.

    Depending on your pngcrush compile options, there may be a different
    number of methods available.  This will find out what they are using the
    program's help text.

    """
    process = subprocess.Popen(['pngcrush', '-h'], stdout=subprocess.PIPE)
    text = process.communicate()[0]
    match = re.search(r'-brute .* \[ (?P<min>\d+) - (?P<max>\d+) \]', text,
                      re.VERBOSE)
    if match is None:
        print >> sys.stderr, "WARNING: Could not determine brute force range."
        return range(11, 125)
    else:
        range_min = int(match.group('min'))
        range_max = int(match.group('max'))
        return range(range_min, range_max+1)


#
# Main function
#

def main():
    """Main program"""
    # Extract any args to pass directly to pngcrush
    opts, args, pngcrush_args = process_args()
    infile, outfile = args
    jobs = opts.jobs

    # If the user requested more jobs than necessary, ignore the empty ones.
    methods = split_rr(get_brute_range(), jobs)
    methods = [x for x in methods if len(x) > 0]

    # Start crushing!
    jobs = [CrushJob(infile, methods[i], pngcrush_args) for i in range(jobs)]
    for j in jobs:
        j.start()

    # Save the best result (the one with the lowest size) as the output.
    result = min([j.result() for j in jobs], key=lambda r: r.size)
    shutil.copyfile(result.path, outfile)


if __name__ == '__main__':
    main()
